from ralium.util.helpers import *
from ralium.util.models import StrPath
from ralium.util.const import *

from pathlib import Path

import mycompat
import inspect
import sys
import os

def get_bundled_filesystem():
    """Returns the bundled filesystem or None."""
    return getattr(sys, SYS_BUNDLE_ATTRIBUTE, None)

def check_bundled():
    """Helper function for checking if the project is bundled."""
    return get_bundled_filesystem() is not None

if check_bundled():
    def get_http_request_handler():
        """Returns a HTTP Server Handler."""
        import ralium.bundle
        return ralium.bundle.BundledHTTPRequestHandler
    
    def check_path_exists(path):
        """Helper function for checking if a path exists."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        return get_bundled_filesystem().exists(path)
    
    def check_path_is_dir(path):
        """Helper function for checking if a path is a directory."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        return StrPath(path) in get_bundled_filesystem().dirs
    
    def read_file_contents(path, encoding="UTF-8"):
        """Helper function for reading files."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)
        
        return get_bundled_filesystem().open(path).decode(encoding)
    
    @_force_args_as_path(absolute=False)
    def create_path(path):
        """
        Helper function for getting file paths.
        If frozen it joins the path with `sys._MEIPASS`.
        Otherwise it just returns the normalized path.
        """
        _raise_if_not_path_type(path)
        return normalize_system_path(path)
else:
    import http.server

    def get_http_request_handler():
        """Returns a HTTP Server Handler."""
        return http.server.SimpleHTTPRequestHandler

    def check_path_exists(path):
        """Helper function for checking if a path exists."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        return path.exists()
    
    def check_path_is_dir(path):
        """Helper function for checking if a path is a directory."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        return path.is_dir()
    
    def read_file_contents(path, encoding="UTF-8"):
        """Helper function for reading files."""
        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        with open(str(path), "r", encoding=encoding) as fp:
            return fp.read()
    
    @_force_args_as_path
    def create_path(path):
        """
        Helper function for getting file paths.
        If frozen it joins the path with `sys._MEIPASS`.
        Otherwise it just returns the normalized path.
        """
        _raise_if_not_path_type(path)

        if mycompat.is_frozen:
            return Path(sys._MEIPASS, path).absolute()

        return path.absolute()

def normalize_webhook_url(path):
    """Normalizes WebHook URL's"""
    return os.path.normpath(f"/{path.lstrip(f'/{BACKSLASH}')}").replace('\\', "/")

@_force_args_as_path(absolute=False)
def normalize_system_path(path):
    """Normalizes System Paths"""
    _raise_if_not_path_type(path)
    
    # Relative paths need to start with a '/' or '\' for windows systems, 
    # this to make bundled projects to work.
    if not path.is_absolute():
        return Path('/' + StrPath(path).lstrip('/'))
    
    return path # pathlib.Path automatically normalizes paths.

is_method = inspect.ismethod
is_function = inspect.isfunction


from ralium.util.common import is_function
from ralium.web.webhook import WebHook

class WebBridge:
    class Collection(dict):
        def add(self, function):
            self[str(function)] = function

    def __init__(self, *functions):
        self.bridges = []

        for function in functions:
            self.new(function)
        
        self.clear = self.bridges.clear
    
    def __str__(self):
        return "".join(self.bridges)
    
    def create_javascript_function(self, name, string):
        return f"function {name}(...calldata){{return pywebview.api.bridge(\"{string}\", calldata)}}"

    def create_javascript_namespace(self, alias, *functions, **named_functions):
        functions = [
            f'{function.__name__}: {self.create_javascript_function("", str(function))}'
            for function in functions
        ]

        functions.extend([
            f'{alias}: {self.create_javascript_function("", str(function))}'
            for alias, function in named_functions.items()
        ])

        return f"let {alias} = {{{','.join(functions)}}};"

    def new(self, obj):
        if is_function(obj):
            return self.bridges.append(self.create_javascript_function(obj.__name__, str(obj)))
        
        if isinstance(obj, WebHook.Namespace):
            return self.bridges.append(self.create_javascript_namespace(obj.alias, **dict(obj.items())))


from .web.webhook import WebHook, create_namespace
from .element import HTMLElement
from .config import WindowConfig
from .window import Window
from .util import __version__
from . import api

from .builtins import (
    getServerPort,
    redirect,
    shutdown,
    refresh,
    getUrl
)

__all__ = [
    "__version__", 
    "create_namespace", 
    "getServerPort", 
    "WindowConfig", 
    "HTMLElement", 
    "redirect", 
    "shutdown", 
    "refresh", 
    "WebHook", 
    "Window", 
    "getUrl",
    "api"
]


from ralium.util.exceptions import WindowRuntimeError
from ralium.util.listener import ClassListener
from ralium.util.common import get_type_name

from ralium.web.webhook import WebHook, create_namespace
from ralium.web.engine import WebEngine

from ralium.navigation import WindowNavigation
from ralium.config import WindowConfig
from ralium.pyhtml import PyHTML
import ralium.builtins

from ralium.util.const import (
    RALIUM_ID_IDENTIFIER_PREFIX,
    HTML_TEMPLATE
)

from ralium.element import (
    create_soup_element, 
    create_ralium_id,
    HTMLElement
)

from bs4 import BeautifulSoup

import webview

@ClassListener
class Window:
    """
    Represents a Ralium GUI Window.

    :param webhooks: A list of WebHooks that handle each URL.
    :param config: Optional configuration for the window.
    """

    def __init__(self, webhooks, config=None):
        self.data = {}
        self.config = config or WindowConfig()
        self.engine = WebEngine(self)
        self.running = False
        self.webhooks = WebHook.Collection(*webhooks)
        self.navigation = WindowNavigation()

        self.webview = webview.create_window(
            html=HTML_TEMPLATE,
            hidden=True,
            js_api=self.engine.create_webview_api(),
            **self.config.as_webview_kwargs(),
            **self.config.other_options
        )

        self.__load_handlers = []
        self.__single_load_handlers = []

        __builtins = create_namespace("ralium", *ralium.builtins.__registry__)

        for webhook in webhooks:
            if self.config.use_builtins:
                webhook.namespaces.append(__builtins)

            webhook.window = self
            webhook.wrap_function_objects()
            webhook.wrap_namespace_objects()
        
        self.engine.start()
    
    def getElementsByClassName(self, className):
        """
        Gets all elements with a specific Class Name.

        :param className: The HTML Class to look for.

        :returns: A list of `HTMLElement` Objects. Returns an empty list if no elements are found.

        :raises WindowRuntimeError: If you call this method before starting the `Window`.
        """

        if not self.running:
            raise WindowRuntimeError("Window must be running before calling the 'getElementsByClassName' method.")

        element_classes = self.webview.evaluate_js(f"Array.from(document.getElementsByClassName('{className}')).map((e) => e.classList);")
        return [HTMLElement(self, classList[0]) for classList in element_classes]
    
    def getElementById(self, id):
        """
        Gets an element with a specific Id.

        :param id: The HTML element Id.
        
        :returns: `HTMLElement` object if found or `None`.
        """

        if self.running:
            element_ralium_id = str(self.webview.evaluate_js(f""" '' + document.getElementById("{id}").classList;"""))
            
            classes = element_ralium_id.split(' ')
            ralium_id = classes[0]

            if ralium_id not in RALIUM_ID_IDENTIFIER_PREFIX:
                for classname in classes:
                    if RALIUM_ID_IDENTIFIER_PREFIX in classname:
                        ralium_id = classname
                        break

            if ralium_id == "null":
                return

            return HTMLElement(self, ralium_id)

        soup = BeautifulSoup(self.webview.html, features="lxml")

        element = soup.select(f'#{id}')
        
        if not element:
            return
        
        ralium_id = element[0].get('class')[0]
        return HTMLElement(self, ralium_id)
    
    def register_load_handler(self, callback, single=False):
        """
        Registers a load handler callback that will be executed when the window is loaded.

        :param callback: The callback function to be executed when the window loads. It must task a single argument.
        :param single: If True, the callback will be executed only once. If False, the callback will be executed every time the window loads.
        :raises TypeError: If the 'single' parameter is not a boolean.
        """

        if not isinstance(single, bool):
            raise TypeError(f"Expected argument 'single' to be a bool object, instead got '{get_type_name(single)}'")
        
        (self.__load_handlers, self.__single_load_handlers)[single].append(callback)

    def load_handler(self, window):
        """
        Handles the loading of the window, executing registered load handler callbacks.

        :param window: The window instance that has been loaded.
        """
        
        self.__single_load_handlers.append(lambda window: window.show())

        def on_single_load():
            for callback in self.__single_load_handlers:
                callback(window)
            
            window.events.loaded -= on_single_load
        
        def on_load():
            for callback in self.__load_handlers:
                callback(window)

        window.events.loaded += on_single_load
        window.events.loaded += on_load
        
    def display(self, url):
        """
        Displays a URL by calling the WebHook attached to the URL.

        :param url: The URL to display.
        """

        if not isinstance(url, str):
            raise TypeError(f"Expected 'url' argumnet to be a 'str' object, instead got '{get_type_name(url)}'")

        self.navigation.location = url
        webhook = self.webhooks.get(self.navigation.location)

        self.engine.bridge.clear()
        self.engine.functions.clear()

        for function in webhook.functions:
            self.engine.bridge.new(function)
            self.engine.functions.add(function)
        
        for namespace in webhook.namespaces:
            self.engine.bridge.new(namespace)

            for function in namespace.values():
                self.engine.functions.add(function)
        
        html = BeautifulSoup(PyHTML(self, webhook).compile(), features="lxml")
        webhook.elements = [create_ralium_id(element) for element in html.body.find_all()]

        html.body.extend([
            create_soup_element(html, "base",   {"href": str(self.engine)}),
            create_soup_element(html, "script", string=str(self.engine.bridge)),
            create_soup_element(html, "style",  string="\n".join(webhook.styles))
        ])

        self.render(html)

    def render(self, html):
        """Write HTML to the DOM."""
        if not self.running:
            self.webview.html = str(html)
            return

        self.webview.load_html(str(html))
    
    def start(self):
        """Starts the Ralium Window and officially displays the GUI."""
        if self.running: return
        
        self.running = True
        webview.start(self.load_handler, self.webview, private_mode=False)
    
    def show(self):
        """Show a Ralium Window."""
        self.webview.show()
    
    def hide(self):
        """Hide a Ralium Window."""
        self.webview.hide()
    
    def shutdown(self):
        """Shutdown a Ralium Window."""
        self.webview.destroy()
        self.engine.close()


import inspect
import logging

# Same logger format as PyInstaller.
# https://github.com/pyinstaller/pyinstaller/blob/develop/PyInstaller/log.py#L35
FORMAT = '%(relativeCreated)d %(levelname)s: %(message)s'

logging.basicConfig(format=FORMAT, level=logging.INFO)

class BasicLogger:
    def __init__(self):
        logger = logging.getLogger("Ralium")
        members = inspect.getmembers(logger, predicate=inspect.ismethod)

        self.disabled = False
        self._methods = []

        for name, method in members:
            self._methods.append(name)
            super().__setattr__(name, method)
    
    def __getattribute__(self, name):
        _methods = super().__getattribute__("_methods")
        disabled = super().__getattribute__("disabled")

        if name in _methods and disabled:
            return lambda *_, **__: None
        
        return super().__getattribute__(name)
    
    def enable(self):
        self.disabled = False
    
    def disable(self):
        self.disabled = True

logger = BasicLogger()
enableLogging = logger.enable
disableLogging = logger.disable


from ralium.util.models import NamedDict
from ralium.util.const import RE_HTML_PATTERN
import copy
import re

class PyHTML:
    __slots__ = ("_html", "_window", "_webhook",)

    def __init__(self, window, webhook):
        self._html = webhook.html
        self._window = window
        self._webhook = webhook
    
    def compile(self):
        """
        Compiles the HTML content of the webhook by processing embedded Ralium tags.

        This method extracts blocks of code enclosed by `{%ralium` and `%}`, processes these blocks,
        executes them, and replaces the original blocks with the resulting HTML content.

        :returns: The compiled HTML content as a string.
        """

        _locals = PyHTML.locals(self._window, self._webhook)
        _blocks = PyHTML.get_enclosed(self._html, "{%ralium", "%}")
        _data = self.process(_blocks)

        for code, block in zip(_data, _blocks):
            outcome = filter(lambda v: v is not None, PyHTML.get_outcome(code, PyHTML.deepcopy(_locals)))
            self._html = self._html.replace(block, "\n".join(outcome))
        
        return self._html
    
    def process(self, blocks):
        """
        Processes blocks of Ralium code to convert HTML tags into Python code.

        This method takes a list of blocks containing Ralium code, processes each block to 
        convert HTML tags into corresponding Python code that appends the tags to a list.

        :param blocks: A list of code blocks to process.

        :returns: A list of processed code blocks as strings.
        """

        data = []

        for block in blocks:
            block = PyHTML.removeaffix(block, "{%ralium", "%}")

            if block.startswith(".inline"):
                data.append(self.process_inline(block.removeprefix(".inline")))
                continue

            html = PyHTML.get_enclosed(block, "{$html", "$}")

            if not html:
                lines = []
                
                for line in block.split("\n"):
                    tags = re.findall(RE_HTML_PATTERN, line)
                    
                    if not tags: 
                        lines.append(line)
                        continue

                    for tag in tags:
                        rep = PyHTML.removeaffix(tag[0], '<">', '</">').replace('"', '\\"')
                        lines.append(line.replace(tag[0], f'__ralium_html__.append("{rep}")'))
                
                data.append("\n".join(lines))
                continue
            
            for tag in html:
                rep = PyHTML.removeaffix(tag, "{$html", "$}").replace('"', '\\"')
                block = block.replace(tag, f'__ralium_html__.append("""{rep}""")')
            
            data.append(block)
        
        return data
    
    def process_inline(self, block):
        """
        Processes an inline Ralium code block to execute it and retrieve the output.

        This method takes an inline block of Ralium code, executes it within a specific local 
        context created from the `Window` and `WebHook` objects, and returns the resulting output.

        :param block: The inline Ralium code block to execute.

        :returns: The output generated by executing the inline code block. Defaults to an empty string.
        """
        
        _locals = PyHTML.locals(self._window, self._webhook)
        _locals["output"] = None

        exec(f"def __ralium_executor__():\n{block}", _locals)
        output = eval("__ralium_executor__()", _locals)

        return f"#[type=inline]{output or ""}"
    
    @staticmethod
    def locals(window, webhook):
        """
        Creates a locals dictionary from a `Window` and `WebHook` object.

        This method generates a dictionary of local variables that includes the functions 
        and namespaces provided by the `WebHook` object, as well as the `Window` object.
        These local variables can be used when executing Python code blocks within the HTML.

        :param window: The `Window` object to expose to the Ralium code block.
        :param webhook: The `WebHook` object containing the functions and namespaces.

        :returns: A dictionary of local variables including the `Window` object, functions, 
        and namespaces from the `WebHook`.
        """

        return {
            "window": window,
            "__ralium_html__": [],
            **{f.__name__: f for f in webhook.functions},
            **{n.alias: NamedDict(n) for n in webhook.namespaces}
        }
    
    @staticmethod
    def deepcopy(locals):
        """
        Creates a deep copy of a locals dictionary without copying the `Window` object.

        This method removes the `Window` object from the dictionary, creates a deep copy of the 
        remaining locals dictionary, and then adds the `Window` object back into the copied 
        dictionary.

        :param locals: The locals dictionary to copy.

        :returns: A deep copy of the locals dictionary with the `Window` object preserved.
        """

        window = locals.pop("window")
        locals_copy = copy.deepcopy(locals)
        locals_copy["window"] = window
        locals["window"] = window

        return locals_copy

    @staticmethod
    def get_outcome(code, locals):
        """
        Executes a Python code block and collects the resulting HTML elements.

        This method executes a given Python code block within a specified local context. If the 
        code block is marked as inline (starting with `#[type=inline]`), it simply returns the 
        code block after removing the inline marker. Otherwise, it executes the code block and 
        collects the resulting HTML elements generated during the execution.

        :param code: The code block to execute.
        :param locals: The dictionary of local variables including functions and namespaces.

        :returns: A list of HTML elements generated by executing the code block, or the code 
                block itself if it is marked as inline.
        """

        if code.startswith("#[type=inline]"):
            return [code.removeprefix("#[type=inline]")]

        exec(f"def __ralium_executor__():\n{code}", locals)
        eval("__ralium_executor__()", locals)

        return locals["__ralium_html__"]

    @staticmethod
    def get_enclosed(text, start, end):
        """
        Extracts all substrings from the given text that are enclosed by the specified start and end markers.

        :param text: The input string to search within.
        :param start: The starting delimiter of the enclosed substrings.
        :param end: The ending delimiter of the enclosed substrings.

        :returns: A list of substrings found within the start and end delimiters.

        :raises SyntaxError: If an ending delimiter is missing for a corresponding start delimiter.
        """

        found = []
        _start = 0

        while (_start := text.find(start, _start + len(start))) != -1:
            _end = text.find(end, _start + len(start))

            if _end == -1:
                raise SyntaxError(f"failed to find ending symbol '{end}'")
            
            _end += len(end)
            found.append(text[_start:_end])
        
        return found

    @staticmethod
    def removeaffix(text, prefix, suffix):
        """
        Removes a prefix and a suffix from a string.

        :param text: The string to process.
        :param prefix: The prefix to remove.
        :param suffix: The suffix to remove.

        :returns: The string without the specified prefix and suffix.
        """

        return text.removeprefix(prefix).removesuffix(suffix)


class ApiError(Exception):
    """The base class for Api related errors."""
    pass

class RegistryNotFoundError(ApiError):
    """
    An error that occurs when the ralium registry was not found.
    Often due to `ralium.api.register()` at the global level of the file
    """
    pass

class ProjectNotFoundError(ApiError):
    """An error that occurs when the specified project directory does not exist."""
    pass

class RoutesNotFoundError(ApiError):
    """An error that occurs when the specified routes directory does not exist."""
    pass

class WebHookError(Exception):
    """The base class for WebHook related errors."""
    pass

class WebHookNotFoundError(WebHookError):
    """
    An error that occurs when trying to render a specific 
    URL that does not have a WebHook attached to it.
    """
    pass

class WebApiError(Exception):
    """The base class for WebApi related errors."""
    pass

class WebBridgeError(WebApiError):
    """
    An error that occurs when a callable is not passed
    to the bridge event wrapper function.
    """
    pass

class SetupError(Exception):
    """The base class for Setup related errors."""
    pass

class WindowError(Exception):
    """The base class for Window related errors."""
    pass

class WindowRuntimeError(WindowError):
    """An error that occurs when the window is not running."""
    pass


from ralium.util.exceptions import WebBridgeError
from ralium.util.common import get_http_request_handler
from ralium.web.bridge import WebBridge
from ralium.element import HTMLElement

from bs4 import BeautifulSoup

import socketserver
import threading
import os

def process_api_calldata(window, calldata):
    if calldata is None:
        return []

    webhook = window.webhooks.get(window.navigation.location)
    new_calldata = []

    for data in calldata:
        if isinstance(data, dict):
            element_id = data.get("className", None)
            local_name = data.get("localName", None)

            if element_id is None or local_name is None or element_id not in webhook.elements: 
                continue

            soup = BeautifulSoup(window.webview.html, "html.parser")
            element = soup.find(local_name, {"class": element_id})

            new_calldata.append(HTMLElement(window, element_id, element))
            continue

        new_calldata.append(data)
    
    return new_calldata

class WebEngine:
    def __init__(self, window):
        self.port, worker = self.initialize_tcp_server()
        
        self._thread = threading.Thread(target=worker, daemon=True)
        self._running = threading.Event()

        self.bridge = WebBridge()
        self.functions = WebBridge.Collection()

        # Strictly for the 'WebviewApi' class created by 'WebEngine.create_webview_api'
        self._window = window

        self.start = self._thread.start

    def __str__(self):
        return f"http://localhost:{self.port}/"

    @property
    def running(self):
        return not self._running.is_set()
    
    def close(self):
        self._running.set()
        self._thread.join(timeout=5)
    
    def create_webview_api(engine):
        class WebviewApi:
            def __init__(self):
                return
            
            # https://github.com/r0x0r/pywebview/issues/1405
            def bridge(self, object, calldata=None):
                if not engine.functions.get(object, False): 
                    return

                calldata = process_api_calldata(engine._window, calldata)

                if calldata:
                    return engine.functions[object](*calldata)
                
                return engine.functions[object]()
        
        return WebviewApi()
    
    def initialize_tcp_server(self):
        httpd = socketserver.TCPServer(("", 0,), get_http_request_handler())
        port = httpd.socket.getsockname()[1]

        def worker():
            while self.running:
                httpd.handle_request()
            os._exit(0)
        
        return port, worker


from ralium.util.common import is_function, is_method
import inspect

class FunctionArgumentInfo:
    def __init__(self, cls, function, result = None, *args, **kwargs):
        self.cls = cls
        self.function = function
        self.result = result
        self.args = list(args)
        self.kwargs = kwargs
    
    def __repr__(self):
        if self.result:
            return f"FunctionArgumentInfo(cls={repr(self.cls)}, function={repr(self.function)}, result='{self.result}', args={self.args}, kwargs={self.kwargs})"
        return f"FunctionArgumentInfo(cls={repr(self.cls)}, function={repr(self.function)}, args={self.args}, kwargs={self.kwargs})"

class ClassListener:
    def __init__(self, cls):
        self.cls = cls
    
    def __call__(self, *args, **kwargs):
        self.cls = self.cls(*args, **kwargs)
        for name, object in inspect.getmembers(self.cls):
            if not (is_function(object) or is_method(object)) or \
                   (name.startswith("__") and name.endswith("__")):
                continue

            setattr(self.cls, name, self.create_listener(object))
        return self.cls
    
    def create_listener(self, function):
        cls = self.cls

        class BroadcastFunction:
            def __init__(self):
                self.__before = {}
                self.__after = {}
            
            def __call__(self, *args, **kwargs):
                info = FunctionArgumentInfo(cls, function, None, *args, **kwargs)

                for f in self.__before.values():
                    f(info)

                info.result = function(*info.args, **info.kwargs)

                for f in self.__after.values():
                    f(info)
                
                return info.result
            
            def subscribe(self, function, is_after = False):
                if is_after:
                    self.__after[str(function)] = function
                    return
                self.__before[str(function)] = function
            
            def unsubscribe(self, function, is_after = False):
                if is_after:
                    del self.__after[str(function)]
                    return
                del self.__before[str(function)]
        
        BroadcastFunction.__name__     = function.__name__
        BroadcastFunction.__qualname__ = function.__qualname__

        return BroadcastFunction()


from pathlib import Path

class StrPath(str):
    """A helper class for creating a `Path` object as a `str` object."""
    def __new__(cls, *args):
        return str(Path(*args)).replace('\\', '/')

class NamedDict(dict):
    def __init__(self, iterable = None):
        for key, value in (iterable or {}).items():
            if isinstance(value, dict):
                self[key] = NamedDict(value)
                continue

            self[key] = value
    
    def __getattr__(self, name):
        if not super().__contains__(name):
            raise AttributeError(f"Object does not contain value '{name}'")
        return super().__getitem__(name)

class NestedNamedDict(NamedDict):
    def __getattr__(self, name, value):
        if not super().__contains__(name):
            super()[name] = NestedNamedDict()
        return super().__getitem__(name, value)


from ralium.util.common import get_type_name
from ralium.util.const import (
    ALL_HTML_ELEMENT_INSTANCE_METHODS, 
    RALIUM_ID_IDENTIFIER_PREFIX
)

import uuid

def create_ralium_id(tag):
    """Creates a new ralium ID to a `BeautifulSoup` element."""
    ralium_id = f"{RALIUM_ID_IDENTIFIER_PREFIX}{uuid.uuid1()}"
    tag['class'] = f"{ralium_id} {' '.join(tag.get('class', []))}".strip()
    return ralium_id

def create_soup_element(soup, tag, attrs=None, **props):
    """
    A helper function for creating new `BeautifulSoup` elements.

    :param soup: A `BeautifulSoup` object.
    :param tag: The name of the element to create.
    :param attrs: A `dict` of HTML attributes to add to the new tag.
    :param props: A keyword `dict` of `BeautifulSoup` attributes to set.

    :returns: The newly created tag.
    """

    element = soup.new_tag(tag)
    attrs = attrs or {}

    if not isinstance(attrs, dict):
        raise TypeError(f"Expected element attributes to be of type dict, instead got '{get_type_name(attrs)}'")

    for name, value in props.items():
        setattr(element, name, value)
    
    for name, value in attrs.items():
        element[name] = value

    return element

class JS:
    class _RawStr(str): pass

    @staticmethod
    def str(__value):
        return f'"{__value}"'
    
    @staticmethod
    def raw(__value):
        return JS._RawStr(__value)

    true = str("true")
    false = str("false")

    @staticmethod
    def bool(__value):
        if __value == "true":
            return True

        if __value == "false":
            return False

    class ClassList(list):
        def __init__(self, element):
            self.element = element
            super().__init__(str(element._eval(f"{element._as_js_str()}.classList")).split(' '))
        
        def add(self, classname):
            if classname in self:
                return

            super().append(classname)
            self.element._eval(f'{self.element._as_js_str()}.classList.add("{classname}")')
        
        def remove(self, classname):
            if classname not in self:
                return
            
            super().remove(classname)
            self.element._eval(f'{self.element._as_js_str()}.classList.remove("{classname}")')
    
HTMLElementAttributes = [
    "_ralium_id", 
    "_ralium_window", 
    "classList"
]

class HTMLElement:
    def __init__(self, window, id):
        self._ralium_id = id
        self._ralium_window = window
        self.classList = JS.ClassList(self)

        if RALIUM_ID_IDENTIFIER_PREFIX in self._ralium_id or not id:
            return

        classes = str(self._eval(f""" '' + document.getElementById("{id}").classList;""")).split(' ')

        if RALIUM_ID_IDENTIFIER_PREFIX in classes[0]:
            self._ralium_id = classes[0]
            return

        for classname in classes:
            if RALIUM_ID_IDENTIFIER_PREFIX in classname:
                self._ralium_id = classname
        
    def __getattr__(self, identifier):
        if identifier in ALL_HTML_ELEMENT_INSTANCE_METHODS:
            def wrapper(*args, **kwargs):
                _prepare = lambda v: (str(v), f'"{v}"')[isinstance(v, str)]
                args_str = [_prepare(arg) for arg in args]
                args_str.extend([f'{_prepare(key)}={_prepare(value)}' for key, value in kwargs.items()])
                args_str = ",".join(args_str)

                return self._eval(f"{self._as_js_str()}.{identifier}({args_str})")
            
            wrapper.__name__     = identifier
            wrapper.__qualname__ = f"HTMLElement.{identifier}"

            return wrapper

        return self._parse(self._eval(f"{self._as_js_str()}.{identifier}"))
    
    def __setattr__(self, identifier, value):
        if identifier in HTMLElementAttributes:
            return object.__setattr__(self, identifier, value)
        
        if isinstance(value, str) and not isinstance(value, JS._RawStr):
            value = JS.str(value)

        self._eval(f"{self._get_js_element()}.{identifier} = {value};")
    
    def _get_js_element(self):
        return f"document.getElementsByClassName('{self._ralium_id}')[0]"
    
    def _as_js_str(self):
        return f"'' + {self._get_js_element()}"
    
    def _eval(self, cmd):
        return self._ralium_window.webview.evaluate_js(cmd)
    
    def _parse(self, value):
        if value in ["true", "false"]:
            return JS.bool(value)
        return value


from ralium.util.common import (
    get_bundled_filesystem,
    normalize_webhook_url,
    get_type_name,
    create_path
)

from ralium.util.models import StrPath
from ralium.util.const import (
    IMAGE_FILE_EXTENSIONS,
    SYS_BUNDLE_ATTRIBUTE
)

from ralium.util.log import logger
from ralium.api import (
    check_project_path,
    check_routes_path
)

from pathlib import Path
from shutil import COPY_BUFSIZE
from http import HTTPStatus

import urllib.parse
import http.server
import importlib
import posixpath
import sys
import os

def reload_ralium():
    """Reloads the all ralium modules."""

    # Get a reference to all modules before reloading them.
    imported_modules = [module for name, module in sys.modules.items() if name.startswith("ralium")]

    for module in imported_modules:
        importlib.reload(module)

# A modified version of http.server.SimpleHTTPRequstHandler
# made compatible with the custom 'FileSystem' class.
class BundledHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def do_GET(self):
        """Serve a GET request."""
        f = self.send_head()
        if f is not None:
            try:
                self.copyfile(f, self.wfile)
            except:
                return

    def send_head(self):
        """
        Common code for GET and HEAD commands.

        This sends the response code and MIME headers.

        Return value is either a file object (which has to be copied
        to the outputfile by the caller unless the command was HEAD,
        and must be closed by the caller under all circumstances), or
        None, in which case the caller has nothing further to do.
        """

        f = None
        path = self.translate_path(self.path)
        ctype = self.guess_type(path)
        # check for trailing "/" which should return 404. See Issue17324
        # The test for this was added in test_httpserver.py
        # However, some OS platforms accept a trailingSlash as a filename
        # See discussion on python-dev and Issue34711 regarding
        # parsing and rejection of filenames with a trailing slash
        if path.endswith("/"):
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None
        try:
            f = get_bundled_filesystem().open(path)
        except FileNotFoundError:
            self.send_error(HTTPStatus.NOT_FOUND, "File not found")
            return None

        try:
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-type", ctype)
            self.send_header("Content-Length", str(len(str(f))))
            self.end_headers()
            return f
        except:
            raise
    
    def translate_path(self, path):
        """
        Translate a /-separated PATH to the local filename syntax.

        Components that mean special things to the local file system
        (e.g. drive or directory names) are ignored.  (XXX They should
        probably be diagnosed.)
        """

        # abandon query parameters
        path = path.split('?',1)[0]
        path = path.split('#',1)[0]
        # Don't forget explicit trailing slash when normalizing. Issue17324
        trailing_slash = path.rstrip().endswith('/')
        try:
            path = urllib.parse.unquote(path, errors='surrogatepass')
        except UnicodeDecodeError:
            path = urllib.parse.unquote(path)
        path = posixpath.normpath(path)
        words = path.split('/')
        words = filter(None, words)
        path = "\\"
        for word in words:
            if os.path.dirname(word) or word in (os.curdir, os.pardir):
                # Ignore components that are not a simple file/directory name
                continue
            path = os.path.join(path, word)
        if trailing_slash:
            path += '/'
        return path

    def copyfile(self, fsrc, fdst):
        """A copy function for `File` objects"""
        fdst_write = fdst.write
        while buf := fsrc[:COPY_BUFSIZE]:
            fdst_write(buf)

class File:
    __slots__ = ("relpath", "content",)

    def __init__(self, relpath, content):
        self.relpath = relpath
        self.content = content
    
    def __str__(self):
        return self.content.decode()

    def __repr__(self):
        return f"ralium.bundle.File(relpath='{StrPath(self.relpath).replace('\\', '/')}', content={repr(self.content)})"
    
class Bundle:
    __slots__ = ("url", "page", "server", "styles",)

    def __init__(self, *, url, page, server, styles):
        self.url = url
        self.page = page
        self.server = server
        self.styles = styles
    
    def __repr__(self):
        return f"ralium.bundle.Bundle(url='{self.url}', page={repr(self.page)}, server={repr(self.server)}, styles=[{''.join([repr(v) for v in self.styles])}])"

class FileSystem:
    __slots__ = ("images", "styles", "bundles", "dirs", "files",)

    def __init__(self, *, images, styles, bundles):
        self.images = images
        self.styles = styles
        self.bundles = bundles

        self.dirs = ['/']
        self.files = {}

        for image in images:
            self.__add_file(image)

        for style in styles:
            self.__add_file(style)
        
        for bundle in bundles:
            self.__add_file(bundle.page)

            if bundle.server is not None:
                self.__add_file(bundle.server)
            
            for style in bundle.styles:
                self.__add_file(style)
    
    def __add_file(self, file):
        # Remove the empty string ''
        segments = StrPath(file.relpath).split('/')[1:]
        segments = [StrPath('/', *segments[:i], v) for i, v in enumerate(segments)]

        self.mkdirs(*segments[:-1])
        self.mkfiles(**{segments[-1]: file.content})

    def open(self, filename):
        if isinstance(filename, Path):
            filename = str(filename)
        elif isinstance(filename, bytes):
            filename = filename.decode()

        try:
            return self.files[StrPath(filename)]
        except KeyError:
            raise FileNotFoundError(f"'{filename}' does not exist.")
    
    def exists(self, path):
        path = StrPath(path)
        if self.files.get(path) is not None:
            return False
        return path in self.dirs

    def mkdirs(self, *dirs):
        self.dirs.extend(filter(lambda v: v not in self.dirs, dirs))
    
    def mkfile(self, filename, data):
        if not isinstance(filename, str):
            raise TypeError(f"Expected filename to be a 'str' object, intead got '{get_type_name(filename)}'")
        
        if not isinstance(data, bytes):
            raise TypeError(f"Expected file content to be a 'bytes' object, instead got '{get_type_name(data)}'.")

        self.files[filename] = data
    
    def mkfiles(self, **files):
        for filename, data in files.items():
            self.mkfile(filename, data)

class PyBundler:
    __slots__ = ("pyfile", "project", "routes", "styles", "images",)

    def __init__(self, pyfile, project):
        self.pyfile = pyfile
        self.project = create_path(project)
        self.routes = self.project/"routes"
        self.styles = self.project/"styles"
        self.images = self.project/"images"

        check_project_path(self.project)
        check_routes_path(self.routes)

    def view(self):
        bundles = []
        shared_images = self.collect(self.images, PyBundler.isimage)
        shared_styles = self.collect(self.styles, PyBundler.isstyle)

        for root, _, files in os.walk(self.routes):
            # Ignore empty directories and __pycache__
            if not files or Path(root).name == "__pycache__":
                logger.info(f"skipping '{root}' ({'empty' if not files else '__pycache__'})")
                continue

            url = normalize_webhook_url(root.removeprefix(str(self.routes)))
            page = None
            server = None
            styles = [*shared_styles]

            for file in files:
                path = Path(root, file).absolute()
                relpath = self.relpath(path)
                filename = path.name

                match filename:
                    case "+page.html":
                        page = File(relpath, self.get_content(path))
                    case "+server.py":
                        server = File(relpath, self.get_content(path))
                    case _:
                        if not PyBundler.isstyle(filename): continue
                        styles.append(File(relpath, self.get_content(path)))
                
                logger.info(f"copying '{path}'")
            
            bundles.append(Bundle(url=url, page=page, server=server, styles=styles))
        
        return [
            b"import ralium.bundle\n",
            b"import sys\n",
            f"setattr(sys, '{SYS_BUNDLE_ATTRIBUTE}', ralium.bundle.FileSystem(\n".encode(),
            f"    images = {shared_images},\n".encode(),
            f"    styles = {shared_styles},\n".encode(),
            f"    bundles = {bundles}\n".encode(),
            b"))\n",
            b"ralium.bundle.reload_ralium()\n\n",
            self.get_content(self.pyfile)
        ]
    
    def relpath(self, path):
        return Path(str(path.absolute()).removeprefix(str(self.project.parent)))

    def collect(self, dir, callback):
        data = []

        for root, _, files in os.walk(dir):
            root = Path(root).absolute()

            for file in files:
                path = root / file
                if not callback(path): continue

                logger.info(f"copying '{path}'")
                data.append(File(self.relpath(path), PyBundler.get_content(path)))
        
        return data

    @staticmethod
    def isstyle(filename):
        return Path(filename).suffix == ".css"
    
    @staticmethod
    def isimage(filename):
        return Path(filename).suffix in IMAGE_FILE_EXTENSIONS
    
    @staticmethod
    def get_content(filename):
        if not Path(filename).exists():
            return
        
        if isinstance(filename, Path):
            filename = str(filename)
        
        with open(filename, "rb") as f:
            return f.read()


# Regex pattern for match single line HTML tags used by `PyHtml`
RE_HTML_PATTERN = r"((<.*>.*</.*>)|(<.*/>))"

# Supported HTML file extensions that `setup` looks for
HTML_FILE_EXTENSIONS = [".htm", ".html"]

# All file extensions that `setup` looks for
FILE_EXTENSIONS = [*HTML_FILE_EXTENSIONS, ".css", ".py"]

# Used for backwards compatibility with Python versions below 3.12. Older versions 
# raise a SyntaxError when you try to include a plain backslash in an f-string, like: f"{'\\'}". 
# Using a constant instead, like: f"{BACKSLASH}" circumvents this problem.
BACKSLASH = "\\"

# This is stored as a constant to make it easy to change in the future. (If Needed)
SYS_BUNDLE_ATTRIBUTE = "bundled"

# Basic HTML Template, used as a fallback for `pywebview`
HTML_TEMPLATE = """<!DOCTYPE html><html><head lang="en"><meta charset="UTF-8"></head><body></body></html>"""

# Supported Image Extensions
IMAGE_FILE_EXTENSIONS = [
    ".apng", ".gif", ".ico", ".cur", ".jpg", ".jpeg", ".webp",
    ".jfif", ".pjpeg", ".pjp", ".png", ".svg", ".icns", ".ico"
]

# The name the ralium api uses for storing exposed functions
RALIUM_API_REGISTRY_IDENTIFIER = "__ralium_registry__"

# The prefix used for creating ralium element ids
RALIUM_ID_IDENTIFIER_PREFIX = "ralium-"

# A list of all HTML Elements
ALL_HTML_ELEMENTS = [
    "a", "abbr", "acronym", "address", "area", "article", "aside", "audio", "b", "base",
    "bdi", "bdo", "big", "blockquote", "body", "br", "button", "canvas", "caption", "center",
    "cite", "code", "col", "colgroup", "data", "datalist", "dd", "del", "details", "dfn", "dialog",
    "dir", "div", "dl", "dt", "em", "embed", "fencedframe", "fieldset", "figcaption", "figure",
    "font", "footer", "form", "frame", "frameset", "h1", "head", "header", "hgroup", "hr", "html",
    "i", "iframe", "img", "input", "ins", "kbd", "label", "legend", "li", "link", "main", "map",
    "mark", "marquee", "menu", "menuitem", "meta", "meter", "nav", "nobr", "noembed", "noframes",
    "noscript", "object", "ol", "optgroup", "option", "output", "p", "param", "picture", "plaintext",
    "portal", "pre", "progress", "q", "rb", "rp", "rt", "rtc", "ruby", "s", "samp", "script", "search",
    "section", "select", "slot", "small", "source", "span", "strike", "strong", "style", "sub", "summary",
    "sup", "table", "tbody", "td", "template", "textarea", "tfoot", "th", "thead", "time", "title", "tr",
    "track", "tt", "u", "ul", "var", "video", "wbr", "xmp"
]

# A list of all methods attached to elements
ALL_HTML_ELEMENT_INSTANCE_METHODS = [
    "after", "animate", "append", "attachShadow", "before", "checkVisibility", "closest",
    "computedStyleMap", "getAnimations", "getAttribute", "getAttributeNames", "getAttributeNode",
    "getAttributeNodeNS", "getAttributeNS", "getBoundingClientRect", "getClientRects",
    "getElementsByClassName", "getElementsByTagName", "getElementsByTagNameNS", "getHTML",
    "hasAttribute", "hasAttributeNS", "hasAttributes", "hasPointerCapture", "insertAdjacentElement",
    "insertAdjacentHTML", "insertAdjacentText", "matches", "prepend", "querySelector",
    "querySelectorAll", "releasePointerCapture", "remove", "removeAttribute", "removeAttributeNode",
    "removeAttributeNS", "replaceChildren", "replaceWith", "requestFullscreen", "requestPointerLock",
    "scroll", "scrollBy", "scrollIntoView", "scrollIntoViewIfNeeded", "scrollTo", "setAttribute",
    "setAttributeNode", "setAttributeNodeNS", "setAttributeNS", "setCapture", "setHTML",
    "setPointerCapture", "toggleAttribute"
]


from pathlib import Path

import inspect

__all__ = ["get_type_name", "_force_args_as_path", "_raise_if_not_path_type"]

def get_type_name(object):
    """Returns the name of an objects type."""
    return type(object).__name__

def __patch_argument(value, absolute):
    if isinstance(value, Path):
        if absolute:
            return value.absolute()
        else:
            return value

    if isinstance(value, str):
        if absolute:
            return Path(value).absolute()
        else:
            return Path(value)

    return value

def _force_args_as_path(absolute=True):
    """
    Converts all arguments of type `str` to a `Path` object.
    Already existing `Path` objects are made absolute if the
    argument `absolute` is set to true.
    """

    _absolute = absolute

    def handler(function):
        def wrapper(*args, **kwargs):
            _patched_args = [__patch_argument(arg, _absolute) for arg in args]
            _patched_kwargs = {key: __patch_argument(value, _absolute) for key, value in kwargs.items()}

            return function(*_patched_args, **_patched_kwargs)
        return wrapper
    
    if inspect.isfunction(absolute):
        _absolute = True
        return handler(absolute)

    return handler

def _raise_if_not_path_type(value):
    """Raises a `TypeError` if `value` is not a `Path` or `str` object."""
    if isinstance(value, str):
        return Path(value)

    if not isinstance(value, Path):
        raise TypeError(f"Expected a 'Path' or 'str' object, instead got '{get_type_name(value)}'")


from ralium.util.common import RALIUM_API_REGISTRY_IDENTIFIER
from ralium.web.webhook import WebHook, create_namespace
from ralium.window import Window

from ralium.util.exceptions import (
    RegistryNotFoundError,
    ProjectNotFoundError,
    RoutesNotFoundError,
    ApiError
)

from ralium.util.common import (
    get_bundled_filesystem, 
    normalize_webhook_url,
    read_file_contents,
    check_path_exists,
    check_path_is_dir,
    check_bundled,
    is_function,
    create_path
)

from pathlib import Path

import inspect
import os

def check_project_path(project):
    if not check_path_exists(project):
        raise ProjectNotFoundError(f"failed to find project directory. ('{project}')")
    
    if not check_path_is_dir(project):
        raise ApiError("Project path must lead to a directory.")

def check_routes_path(routes):
    if not check_path_exists(routes):
        raise RoutesNotFoundError("Project directory must contain a routes directory.")
    
    if not check_path_is_dir(routes):
        raise ApiError("Project routes path must lead to a directory.")

def collect_project_shared_styles(project):
    styles_path = project/"styles"
    styles = []

    if not check_path_exists(styles_path) or not check_path_is_dir(styles_path):
        return []

    for root, _, files in os.walk(styles_path):
        styles.extend([Path(root, file).absolute() for file in files if Path(file).suffix == ".css"])
    
    return styles

def get_server_registry_contents(source):
    globals = {}
    exec(read_file_contents(source), globals)
    return globals.get(RALIUM_API_REGISTRY_IDENTIFIER, [])

def get_server_registry_objects(source):
    functions = []
    namespaces = []

    for obj in get_server_registry_contents(source):
        if isinstance(obj, WebHook.Namespace):
            namespaces.append(obj)
            continue

        if is_function(obj):
            functions.append(obj)
    
    return functions, namespaces

if check_bundled():
    def collect_webhooks(_, __):
        filesystem = get_bundled_filesystem()

        webhooks = []
        shared_styles = [str(style) for style in filesystem.styles]

        for bundle in filesystem.bundles:
            styles = [*shared_styles, *[str(style) for style in bundle.styles]]

            functions = []
            namespaces = []

            if bundle.server is not None:
                functions, namespaces = get_server_registry_objects(bundle.server.relpath)
            
            webhooks.append(WebHook(bundle.url, str(bundle.page), styles, functions, namespaces))
        
        return webhooks
else:
    def collect_webhooks(project, routes):
        webhooks = []

        shared_styles = collect_project_shared_styles(project)
        root_prefix = str(project/"routes")

        for root, _, files in os.walk(routes):
            # Skip empty directories
            if not files or Path(root).name == "__pycache__":
                continue

            url = normalize_webhook_url(root.removeprefix(root_prefix))
            html = None
            styles = [*shared_styles]
            functions = []
            namespaces = []

            for file in files:
                file = Path(file)
                path = Path(root, file)

                if file.name == "+page.html":
                    html = path
                elif file.suffix == ".css":
                    styles.append(path)
                elif file.name == "+server.py":
                    functions, namespaces = get_server_registry_objects(path)
            
            if html is None:
                raise ApiError(f"All routes must contain a '+page.html' file. ('{root}')")
            
            webhooks.append(WebHook(url, html, styles, functions, namespaces))
    
        return webhooks

class Module:
    """
    A container, typically used for module methods, that wraps functions and named functions that don't need a window argument.

    :param functions: A list of functions or methods to wrap.
    :param named_functions: A dictionary of functions with names different from their current ones to wrap.
    """

    # Flag attribute used by the 'WebHook.Namespace' class.
    _module_api_class = True
    
    def __init__(self, *functions, **named_functions):
        self.functions = [wrap(function) for function in functions]
        self.named_functions = {name: wrap(function) for name, function in named_functions.items()}

def create_window(project, config = None):
    """
    Creates a window from a specific directory structure.

    :param project: The project directory.
    :param config: The window configuration.

    :returns: A `Window` object.

    :raises ApiError: If the `project` or `routes` paths are not directories.
    :raises ProjectNotFoundError: If the project directory was not found.
    :raises RoutesNotFoundError: If the routes directory was not found.
    """

    project = create_path(project)
    routes = project/"routes"

    check_project_path(project)
    check_routes_path(routes)

    webhooks = collect_webhooks(project, routes)

    return Window(webhooks, config)

def register():
    """
    Creates a registry list to expose functions and namespaces for use within JavaScript.

    :returns: A decorator for exposing Python functions to JavaScript.
    """

    globals = inspect.currentframe().f_back.f_globals
    globals[RALIUM_API_REGISTRY_IDENTIFIER] = []

    def wrapper(function):
        if not is_function(function):
            raise ApiError("Expose decorator must be called on a 'function' object.")

        if function not in globals[RALIUM_API_REGISTRY_IDENTIFIER]:
            globals[RALIUM_API_REGISTRY_IDENTIFIER].append(function)
        
        def echo(*args, **kwargs):
            return function(*args, **kwargs)
        
        echo.__name__     = function.__name__
        echo.__qualname__ = function.__qualname__
        echo.__function__ = function

        return echo

    return wrapper

def namespace(alias, *functions, **named_functions):
    """
    Adds a namespace to the server registry.

    :param alias: The name the namespace is accessed by.
    :param functions: A list of functions to add.
    :param named_functions: A dictionary of functions to add.

    :returns: A `WebHook.Namespace` object.

    :raises RegistryNotFoundError: If `ralium.api.register` is not called at the global level.

    The alias of the regularly listed functions will be the value
    of the `__name__` property of the function objects.

    The alias of the named function dictionary will be the keyword part.
    """

    registry = inspect.currentframe().f_back.f_globals.get(RALIUM_API_REGISTRY_IDENTIFIER)

    if registry is None:
        raise RegistryNotFoundError(
            "No active registry found in the global space.",
            "Ensure that 'ralium.api.register' has been called to initialize the registry.",
            "('ralium.api.register' must be called at the global level of the file)"
        )
    
    for function in [*functions, *named_functions.values()]:
        if hasattr(function, "__function__"):
            function = getattr(function, "__function__")
        
        if function in registry:
            raise ApiError(f"Cannot add already exposed function '{function.__name__}' to a namespace.")
    
    registry.append(create_namespace(alias, *functions, **named_functions))

def wrap(function):
    """
    Wraps a function to accept a 'window' argument, typically for module methods.
    The wrapped function does not get the `Window` object passed to it.

    :param function: A function or method.

    :returns: A decorator for the wrapped function.
    """

    def wrapper(_, *args, **kwargs):
        return function(*args, **kwargs)
    
    wrapper.__name__     = function.__name__
    wrapper.__qualname__ = function.__qualname__
    
    return wrapper


from PyInstaller.__main__ import run as create_executable

from ralium.util.exceptions import SetupError
from ralium.util.common import check_bundled, get_type_name
from ralium.util.const import FILE_EXTENSIONS
from ralium.util.log import logger
from ralium.bundle import PyBundler
from ralium.util import __version__

from pathlib import Path
import subprocess
import mycompat
import sys
import os

def __create_exe_str_error(name, var):
    return TypeError(f"Expected executable {name} to be of type 'str', instead got '{get_type_name(var)}'")

def __create_add_data_arg(src, dst):
    return f'--add-data={src}{os.pathsep}{dst}'

def collect_project(project):
    """
    Collects all necessary files from a Ralium Project used within the executable main Python file.

    :param project: The path to the project folder.
    
    :returns: A `dict` where the keys contain the absolute paths and the values represent the relative paths.
    """

    files = {}
    project = Path(project).absolute()
    routes = project/"routes"
    styles = project/"styles"
    images = project/"images"
    
    if os.path.exists(routes):
        for root, _, files in os.walk(routes):
            dest = Path(project.name, root.removeprefix(str(project)))
            files.update({str(Path(root, file).absolute()): dest for file in files if Path(file).suffix in FILE_EXTENSIONS})
    
    if os.path.exists(styles):
        for root, _, files in os.walk(styles):
            dest = Path(project.name, root.removeprefix(str(project)))
            files.update({str(Path(root, file).absolute()): dest for file in files if Path(file).suffix == ".css"})
    
    if os.path.exists(images):
        for root, _, files in os.walk(images):
            dest = Path(project.name, root.removeprefix(str(project)))
            files.update({str(Path(root, file).absolute()): dest for file in files})
    
    return files

def bundle(
    pyfile,
    project,
    filename = None,
    distpath = None
):
    """
    Bundles a Ralium project to a single python file.

    :param pyfile: The python file to compile.
    :param project: The path to the project directory.
    :param filename: The name to set for the bundled file.
    :param distpath: The directory to create the bundle in. (Default: `dist`)
    
    :returns: The path to the bundled file created.

    :raises TypeError: If the incorrect types are provided for certain arguments.
    :raises SetupError: If `project` is None.
    :raises FileNotFoundError: If the `pyfile` or the `project` path does not exist.
    """

    if project is None:
        raise SetupError("Cannot bundle project without a project directory.")

    distpath = Path(distpath or "dist").absolute()

    if not os.path.exists(distpath):
        os.mkdir(distpath)
        logger.info("created %s", distpath)

    base, ext = os.path.splitext(os.path.basename(pyfile))
    filename = os.path.join(distpath, f"{filename or base}.bundle{ext}")

    logger.info("Bundling '%s' with project '%s'", pyfile, os.path.abspath(project))

    code = PyBundler(pyfile, project).view()

    # PyInstaller 6.6.0 and greater require these modules on Windows.
    # If you compile the program without these lines added to the file.
    # An ImportError will occur in which the pywin32-ctypes module needs to be installed.
    # Adding these lines as imports before everything else fixes this problem.
    if mycompat.is_win:
        code.insert(0, b"from win32ctypes.pywin32 import pywintypes\n")
        code.insert(1, b"from win32ctypes.pywin32 import win32api\n")

    with open(filename, "wb") as f:
        f.writelines(code)

    logger.info("wrote %s", filename)
    logger.info("Finished Bundling")

    return os.path.abspath(filename)

def setup(
    pyfile,
    name = None,
    icon = None,
    bundle = False,
    project = None,
    onefile = True,
    noconsole = True,
    bundle_dist = None,
    optimize = None,
    pyi_args = None,
    use_subprocess = False,
):
    """
    Compiles a Ralium project to an executable using PyInstaller.

    :param pyfile: The python file to compile.
    :param name: Display name for the executable.
    :param icon: Display icon for the executable.
    :param bundle: Bundles all of the HTML, CSS, python and image files into one executable. (Requires a project directory)
    :param project: Directory of the project.
    :param onefile: Creates the executable as a standalone file.
    :param noconsole: Prevents a console from being displayed.
    :param bundle_dist: The directory name Ralium will use for bundling projects. (Default: `dist`)
    :param optimize: Set the `PYTHONOPTIMIZE` level or the PyInstaller `--optimize` flag. (Default: `0`)
    :param pyi_args: Extra parameters for PyInstaller to use.
    :param use_subprocess: Use `PyInstaller` through a subprocess.

    :raises TypeError: If the name or icon is not a `str` or `None`.
    :raises SetupError: If `bundle` is True while the `project` directory is None or does not exist.
    :raises RuntimeError: If this function is called within an already compiled executable file.
    :raises FileNotFoundError: If a certain file path doesn't exist.
    """

    if not isinstance(pyfile, (Path, str, bytes,)):
        raise TypeError(f"Expected 'pyfile' to be a 'Path', 'str', or 'bytes' object, instead got '{get_type_name(pyfile)}'")

    logger.info("Ralium: %s", __version__)

    if mycompat.is_frozen:
        raise RuntimeError("Ralium 'setup' cannot be ran from an executable file.")
    elif check_bundled():
        raise RuntimeError("Ralium 'setup' cannot be ran from a bundled file.")

    if not Path(pyfile).exists():
        raise FileNotFoundError(f"Failed to find python file '{pyfile}'")
    
    if optimize not in [None, *range(0, 3)]:
        raise ValueError("The optimization level must be the value of 0, 1, or 2.")

    args = [pyfile, *(pyi_args or [])]

    if name is not None:
        if not isinstance(name, str):
            raise __create_exe_str_error("name", name)
        
        args.append(f"--name={name}")
    
    if icon is not None:
        if not isinstance(icon, str):
            raise __create_exe_str_error("icon", icon)

        if not os.path.exists(icon):
            raise FileNotFoundError(f"Failed to find icon file with path: '{icon}'")
        
        args.append(f"--icon={icon}")
    
    if onefile:
        args.append(f"--onefile")
    
    if noconsole:
        args.append(f"--noconsole")
    
    if bundle:
        args[0] = globals()["bundle"](
            pyfile=pyfile, 
            project=project,
            filename=name,
            distpath=bundle_dist
        )
    elif project:
        for src, dst in collect_project(project).items():
            args.append(__create_add_data_arg(src, dst))

    if use_subprocess:
        optimize_flag = ""

        match optimize:
            case 1: optimize_flag = "-O"
            case 2: optimize_flag = "-OO"
        
        args = [
            sys.executable,
            "-m",
            "PyInstaller",
            *args
        ]

        if optimize_flag:
            args.insert(1, optimize_flag)
        
        return subprocess.run(args) and None
    else:
        args.append(f"--optimize={optimize or 0}")
    
    create_executable(args)


from ralium.element import JS
import webbrowser

__registry__ = []

def builtin(function):
    if function not in __registry__:
        __registry__.append(function)
    
    return function

@builtin
def refresh(window):
    window.display(window.navigation.location)

@builtin
def redirect(window, url):
    window.display(url)

@builtin
def shutdown(window):
    window.shutdown()

@builtin
def getUrl(window, element = None):
    element.innerHTML = JS.str(window.navigation.location)
    return window.navigation.location

@builtin
def getServerPort(window, element = None):
    element.innerHTML = window.engine.port
    return window.engine.port

@builtin
def openBrowserTab(window, path):
    return webbrowser.open_new_tab(path)


from ralium.util.common import (
    _raise_if_not_path_type,
    normalize_system_path,
    normalize_webhook_url,
    read_file_contents,
    check_path_exists,
    get_type_name,
    is_function
)

from ralium.util.exceptions import (
    WebHookNotFoundError, 
    WebHookError
)

from collections.abc import Iterable
from pathlib import Path

def create_namespace(alias, *functions, **named_functions):
    """
    Adds a `WebHook.Namespace` object to JavaScript.

    :param alias: The name the namespace is accessed by.
    :param functions: A list of functions to add.
    :param named_functions: A dictionary of functions to add.

    :returns: A `WebHook.Namespace` object.

    The alias of the regularly listed functions will be the value
    of the `__name__` property of the function objects.

    The alias of the named function dictionary will be the keyword part.
    """
    return WebHook.Namespace(alias, *functions, **named_functions)

class WebHook:
    """
    A WebHook Object that contains information for handling a specific URL.

    :param url: The URL that the WebHook handles.
    :param html: A file path or raw text of the HTML to render.
    :param styles: A list or tuple of paths or raw css content.
    :param functions: Functions to expose to JavaScript.
    :param namespaces: Namespaces to expose to JavaScript.
    :param homepage: If this WebHook is a homepage, it will be the fallback page if something goes wrong with the `Navigation` handler.
    :param encoding: The file encoding of the HTML and CSS files.
    """

    class Function:
        def __new__(cls, function, window):
            def wrapper(*args, **kwargs):
                return function(window, *args, **kwargs)
            
            wrapper.__name__     = function.__name__
            wrapper.__qualname__ = function.__name__
            wrapper.__webhook__  = True

            return wrapper

    class Namespace(dict):
        def __init__(self, alias, *functions, **named_functions):
            self.alias = alias

            self.add_functions(*functions)
            self.add_named_functions(**named_functions)
        
        def add(self, name, function):
            if not is_function(function):
                raise TypeError(f"Expected a 'function' object for namespace, instead got '{get_type_name(function)}'")
            
            self.__setitem__(name, function)
        
        def add_functions(self, *functions):
            for function in functions:
                self.add(function.__name__, function)
        
        def add_named_functions(self, **functions):
            for name, function in functions.items():
                self.add(name, function)
    
    class Collection(dict):
        """A helper class for storing `WebHook` objects."""
        
        def __init__(self, *webhooks):
            for webhook in webhooks:
                if not isinstance(webhook, WebHook):
                    raise TypeError("WebHook.Collection can only contain 'WebHook' objects.")

                self[webhook.url] = webhook

        def __iter__(self):
            return self.values().__iter__

        def __repr__(self):
            return f"{{{', '.join([repr(webhook) for webhook in self.values()])}}}"
        
        def get(self, url):
            webhook = super().get(url, None)

            if webhook is None:
                raise WebHookNotFoundError(f"Failed to find WebHook for the url '{url}'")

            return webhook

    def __init__(self, url, html, styles=None, functions=None, namespaces=None, encoding="UTF-8"):
        # Needs to be initialized before html, and styles
        self.encoding = encoding

        self.url = normalize_webhook_url(url)
        self.html = html
        self.styles = styles or []

        self.window = None
        self.elements = []
        self.functions = functions or []
        self.namespaces = namespaces or []

    def __repr__(self):
        return f"WebHook(url='{self.url}')"

    @property
    def html(self):
        return self._html
    
    @property
    def styles(self):
        return self._styles
    
    @html.setter
    def html(self, new_html):
        if isinstance(new_html, bytes):
            new_html = new_html.decode(self.encoding)

        if isinstance(new_html, Path):
            self._html = WebHook.load_html(new_html, self.encoding)
            return
        elif isinstance(new_html, str):
            # Assume if the starting line is <!DOCTYPE html> that it is raw HTML.
            if new_html.startswith("<!DOCTYPE html>"):
                self._html = new_html
                return

            path = Path(new_html)

            if check_path_exists(path):
                self._html = WebHook.load_html(path, self.encoding)
                return
            
            self._html = new_html
            return
        
        raise TypeError(f"WebHook HTML property expected 'Path', 'str' or 'bytes' object, instead got '{get_type_name(new_html)}'")
    
    @styles.setter
    def styles(self, new_styles):
        if not isinstance(new_styles, Iterable):
            raise TypeError(f"WebHook styles property expected an iterable, instead got '{get_type_name(new_styles)}'")
        
        _styles = []

        for style in new_styles:
            if isinstance(style, bytes):
                style = style.decode(self.encoding)

            if isinstance(style, str):
                path = Path(style)

                if not check_path_exists(path):
                    _styles.append(style)
                    continue

                _styles.append(WebHook.load_style(path, self.encoding))
                continue

            if isinstance(style, Path):
                _styles.append(WebHook.load_style(style, self.encoding))
                continue
            
            raise TypeError(
                f"Styles {get_type_name(new_styles)} contained unexpected type '{get_type_name(new_styles)}',"
                "expected 'Path', 'str', or 'bytes' object."
            )

        self._styles = _styles
    
    @staticmethod
    def requires_window(function):
        """Raises if a method is called without `WebHook.window` being set."""
        def wrapper(self, *args, **kwargs):
            if not self.window:
                raise WebHookError("cannot call '{function.__name__}', the window has not been set.")
            return function(self, *args, **kwargs)
        return wrapper

    @requires_window
    def wrap_function_objects(self):
        """Wraps all functions with `WebHook.Function`"""
        self.functions = [WebHook.Function(function, self.window) for function in self.functions if not hasattr(function, "__webhook__")]
    
    @requires_window
    def wrap_namespace_objects(self):
        """Wraps all `WebHook.Namespace` objects child functions with `WebHook.Function`"""
        for namespace in self.namespaces:
            for key, function in namespace.items():
                if hasattr(function, "__webhook__"):
                    continue
                namespace[key] = WebHook.Function(function, self.window)

    @staticmethod
    def _loads(path, encoding="UTF-8", file_type=None):
        """
        Get the contents of a HTML or CSS file from a `Path` object.
        
        :param path: A `Path` or `str` object of the file to load.
        :param encoding: The encoding to use when reading the file.
        :param file_type: Used for more specific errors.

        :raises TypeError: If the path is not a `Path` or `str` object.
        :raises FileNotFoundError: If the specified path does not exist.
        """

        path = normalize_system_path(path)
        _raise_if_not_path_type(path)

        if not check_path_exists(path):
            raise FileNotFoundError((
                f"Failed to find {file_type} file. ('{path}')", 
                f"Failed to find file. ('{path}')"
            )[file_type is None])
        
        return read_file_contents(path, encoding)
    
    load_html = staticmethod(lambda path, encoding="UTF-8": WebHook._loads(path, encoding, "HTML"))
    load_style = staticmethod(lambda path, encoding="UTF-8": WebHook._loads(path, encoding, "CSS"))